"""
This type stub file was generated by pyright.
"""

from email.message import Message

"""Extra packaging metadata used by OZI."""
sspace = ...
dcolon = ...
classifier = ...
pep639_headers = ...
pep639_headers_md = ...
license_expression = ...
license_file = ...
extra_classifiers_comment = ...
def parse_extra_pkg_info(pkg_info: Message) -> tuple[dict[str, str], str | None]:
    """Get extra Classifiers tacked onto the PKG-INFO payload by OZI."""
    ...

