"""
This type stub file was generated by pyright.
"""

"""``ozi-new`` console application.
"""
parser = ...
subparser = ...
interactive_parser = ...
PROJECT_METADATA_HELP = ...
project_parser = ...
interactive_defaults = ...
required = ...
optional = ...
defaults = ...
ozi_defaults = ...
ozi_required = ...
output = ...
