"""
This type stub file was generated by pyright.
"""

from typing import Optional
from typing import Union

from .exceptions_types import ValidatedEmail

def validate_email(email: Union[str, bytes], /, *, allow_smtputf8: Optional[bool] = ..., allow_empty_local: bool = ..., allow_quoted_local: Optional[bool] = ..., allow_domain_literal: Optional[bool] = ..., check_deliverability: Optional[bool] = ..., test_environment: Optional[bool] = ..., globally_deliverable: Optional[bool] = ..., timeout: Optional[int] = ..., dns_resolver: Optional[object] = ...) -> ValidatedEmail:
    """
    Given an email address, and some options, returns a ValidatedEmail instance
    with information about the address if it is valid or, if the address is not
    valid, raises an EmailNotValidError. This is the main function of the module.
    """
    ...

