"""
This type stub file was generated by pyright.
"""

from __future__ import annotations

import json
import os
from argparse import Namespace
from pathlib import Path
from typing import TYPE_CHECKING
from typing import NoReturn

from jinja2 import Environment
from ozi_spec import METADATA  # pyright: ignore
from ozi_templates import load_environment
from ozi_templates.filter import underscorify
from tap_producer import TAP

from ozi_core.fix.missing import report
from ozi_core.fix.parser import parser
from ozi_core.fix.rewrite_command import Rewriter
from ozi_core.new.validate import valid_copyright_head

"""ozi-fix: Project fix script that outputs a meson rewriter JSON array."""
if TYPE_CHECKING:
    ...
def main() -> NoReturn:
    """Main ozi.fix entrypoint."""
    ...

