"""
This type stub file was generated by pyright.
"""

from enum import Enum
from enum import unique
from typing import Any

from ozi_spec import METADATA

def checkbox(checked: bool) -> str:
    ...

@unique
class MenuButton(Enum):
    """Non-composable menu action enum."""
    ADD = ...
    BACK = ...
    EDIT = ...
    EXIT = ...
    MENU = ...
    METADATA = ...
    OK = ...
    OPTIONS = ...
    PROMPT = ...
    REMOVE = ...
    RESET = ...


def main_menu(project: Any, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | bool, dict[str, list[str]], dict[str, str]]:
    ...

def options_menu(project: Any, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | bool, dict[str, list[str]], dict[str, str]]:
    ...

def edit_menu(project: Any, output: dict[str, list[str]], prefix: dict[str, str]) -> tuple[None | list[str] | bool, dict[str, list[str]], dict[str, str]]:
    ...

